/* KWP2000.C
 * 
 * Provides C routines for accessing ISO-14230 box providing KWP2000 features.
 * Front-end allows:
 *
 *           (a) Fast init w/keyword retrieval
 *           (b) Slow init w/keyword retrieval
 *           (c) ISO-14230 box status code query
 *           (d) Send message capability in init mode
 *           (e) Send message capability in run mode
 *           (f) receive message capability in run mode
 *           (g) Most recent keyword retrieval
 *           (h) Clear ISO-14230 box buffers
 *           (i) Voltage level setting
 *	     (j) Revision level query
 *	     (k) Go directly to RUN mode
 *
 *
 * Usage:  KWP2000 [P<comport>]
 *
 * Optional argument specifies which serial COM port the ISO-14230 box is
 * connected to.  Default is COM1.  Example:  KWP2000 P2 (for COM2) 
 *
 * 3/09/02 Prune
 */
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <signal.h>
#include <dos.h>
#include <time.h>
#include "nissan.h"
#include "isoapp.h"
#include "isobox.h"
#include "rs232.h"

#define CANDISP(c) (c&&((c<7)||(c>13))&&((c<0x1C)||(c>0x1F))&&((c&127)!=127))

typedef enum {
    UNDETERMINED_MODE=0,
    INIT_MODE,
    RUN_MODE,
    RUN_SEND,
    RUN_RECEIVE,
    RUN_RECEIVE_SEQ,
    RUN_PARM,
    RUN_SET_PARM
} ModeEnum;

FILE *fp;
time_t timestamp;
ModeEnum mode;
ParmEnum ParmNum;
unsigned char EnableTesterPresent;
unsigned char incval;
unsigned int resp;
unsigned char seqstart;

/* Necessary Subroutine */
void clean(int error_code);

main(int argc, char **argv)
{
unsigned char quit,printmenu,stringpos,*pp,val;
unsigned int i;
int data,res;
    quit=0; comport=0; EnableTesterPresent=0;
    while (--argc>0) { i=atoi((pp=strupr(*++argv))+1); switch (*pp) {
      	case '/': if (pp[1]!='?') { printf("Arg: no slash needed"); exit(-1); }
	case '?': printf("\rArgs:  Pdd   comport \n"
			);
			exit(0);
    	case 'P': printf("COM%d\n",(comport=i-1)+1); case 0: break;
	default: printf("UNKNOWN "); if (CANDISP(*pp)) { printf("\'%c\'",*pp);
	    printf(" FROM ARG "); } printf("\"%s\"",pp);
    }} setverify(1); signal(SIGINT,clean);
    /* printf("%lu bytes available for allocation\n",farcoreleft()); */
    mode=UNDETERMINED_MODE;
    LoadParameters();
    if (!InitComm(comport,BAUDRATE)) {
    	printf("Could not initialize COM port.\n");
    	clean(-3);
    }
    while (!quit) {
	switch (mode) {
	    case UNDETERMINED_MODE:
		if (modemstatus() & DSR) {
		    printf("Already in RUN mode.\n");
		    printmenu=1;
		    mode=RUN_MODE;
		} else {
		    mode=INIT_MODE;
		} break;
	    case INIT_MODE:
		handshake(RTS);
		printf("\nINIT MODE MENU\n\n"
		    "(1) Query Status \n"
		    "(2) Set Line Voltage \n"
		    "(3) Query Last Keyword \n"
		    "(4) Fast Init \n"
		    "(5) Slow Init \n"
		    "(6) Go to RUN mode \n"
		    "(7) Clear Buffers \n"
		    "(8) Send Message \n"
		    "(9) Query Firmware Revision Level \n"
		    "(10) Quit \n\nEnter Choice (1-10): ");
		gets(string); 
		i=atoi(string);
		printmenu=1; /* in case we goto RUN mode */
		switch (i) {
		    case 1: QueryStatus(); break;
		    case 2: SetLineVoltage(); break;
		    case 3: QueryKeyword(); break;
		    case 4: if (FastInit()>=0) mode=RUN_MODE; break;
		    case 5: if (SlowInit()>=0) mode=RUN_MODE; break;
		    case 6: 
			if (SendCommand('G')==0) {
			    if (RecvInitResp(&resp,RESP_STATUS)==0) {
		   	        handshake(RTS|DTR);	
				delay(1000);
				mode=RUN_MODE;
			    }
			} 
			break;
		    case 7: ClearBuffers(); break;
		    case 8: SendMessage(); break;
		    case 9: QueryRevision(); break;
		    case 10: quit=1; break;
		} break;
	    case RUN_MODE:
		MyClearRXBuffer();
		if ((time(NULL)-timestamp)>=2UL) { /* 2 seconds passed? */
		    TesterPresent();     /* send Tester Present message */
		    time(&timestamp);    /* to keep modules in RUN mode */
		}
		if (!(modemstatus() & DSR)) {
		    printf("\n\nModule has reverted to INIT mode.\n");
		    mode=INIT_MODE;
		    break;
		}
		if (printmenu) {
		    printmenu=0;
		    stringpos=0;
		    printf("\nRUN MODE MENU\n\n"
		        "(1) Receive Data on K-Line\n"
			"(2) Receive Sequential Data on K-Line\n"
		        "(3) Send Sequential Data on K-Line\n" 
			"(4) Return to INIT mode\n"
			"(5) Edit Parameters\n"
			"(6) Quit\n\nEnter Choice: ");
		}
		if (kbhit()) {
		    string[stringpos]=getch();
		    switch (string[stringpos]) {
			case 0: getch(); break; /* special char - ignore */
			case 8:  /* backspace */
			    if (stringpos) {
				printf("\x08 \x08");
				stringpos--;
			    } break;
			case 13: case 10: /* carriage return */
			    string[stringpos]=0;
			    printf("\n");
			    i=atoi(string);
			    switch (i) {
				case 1: mode=RUN_RECEIVE; printmenu=1; break;
				case 2: mode=RUN_RECEIVE_SEQ; printmenu=1; seqstart=0; break;
				case 3: mode=RUN_SEND; printmenu=1; break;
				case 4: mode=INIT_MODE; break;
				case 5: mode=RUN_PARM; printmenu=1; break;
				case 6: quit=1; break;
			        default: 
				    printmenu=1;
				    printf("\nInvalid option.\n");
			    } break;
			default:
			    if (CANDISP(string[stringpos])) {
				printf("%c",string[stringpos]);
			        if (stringpos<50) stringpos++;
			    }
		    }
		} break;
	    case RUN_SEND:
		MyClearRXBuffer();
#ifdef NOT_NECESSARY
		if ((time(NULL)-timestamp)>=2UL) { /* 2 seconds passed? */
		    TesterPresent();     /* send Tester Present message */
		    time(&timestamp);    /* to keep modules in RUN mode */
		}
#endif
		if (!(modemstatus() & DSR)) {
		    printf("\n\nModule has reverted to INIT mode.\n");
		    mode=INIT_MODE;
		    break;
		}
		MyWriteChar(incval++);
		if (!incval) printf(".");
#ifdef NISSAN_DEF
		if (printmenu) {
		    printmenu=0;
		    stringpos=0;
		    printf("\nRUN MODE SEND MESSAGE\n\n"
			"1. Read Freeze Frame Data\n"
			"2. Clear Diagnostic Trouble Codes\n"
			"3. Read ECU Identification\n"
			"4. Read Data by Local Identifier\n"
			"5. Read Data by Common Identifier\n"
			"6. Input Output Control by Local Identifier\n"
			"7. Write Data by Local Identifier\n"
			"8. Read Diagnostic Trouble Codes with Option Byte\n"
			"9. Read Time of Diagnostic Trouble Codes\n"
			"10. Simply Dynamically Define Local ID\n"
			"11. Supplier Command (Valve Activation)\n"
		        "12. Clear Diagnostic Information\n"
 		        "13. EEPROM Dump\n"
			"14. Read Software Version\n"
			"15. Quit to Main RUN Mode Menu\n"
			"\nEnter Choice (1-15): ");
		} 
		if (kbhit()) {
		    string[stringpos]=getch();
		    switch (string[stringpos]) {
			case 0: getch(); break; /* special char - ignore */
			case 8:  /* backspace */
			    if (stringpos) {
				printf("\x08 \x08");
				stringpos--;
			    } break;
			case 13: case 10: /* carriage return */
			    string[stringpos]=0;
			    printf("\n");
			    printmenu=1;
			    i=atoi(string);
			    if ((i>=1)&&(i<=NUMBER_OF_PARAMETERS)) {
				mode=RUN_RECEIVE;
			    }
			    switch (i) {
				case 1: ReadFreezeFrameData(); break;
				case 2: ClearDiagnosticTroubleCodes(); break;
				case 3: ReadECUIdentification(); break;
				case 4: ReadDataByLocalIdentifier(); break;
				case 5: ReadDataByCommonIdentifier(); break;
				case 6: InputOutputControlByLocalIdentifier(); break;
				case 7: WriteDataByLocalIdentifier(); break;
				case 8: ReadDiagnosticTroubleCodesWithOptionByte(); break;
				case 9: ReadTimeOfDiagnosticTroubleCodes(); break;
				case 10: SimplyDynamicallyDefineLocalID(); break;
				case 11: ValveActivation(); break;
				case 12: ClearDiagnosticInformation(); break;
				case 13: EEPROMDump(); break;
				case 14: ReadSoftwareVersion(); break;
				case 15: mode=RUN_MODE; break;
			        default: printf("\nInvalid option.\n");
			    } break;
			default:
			    if (CANDISP(string[stringpos])) {
				printf("%c",string[stringpos]);
			        if (stringpos<50) stringpos++;
			    }
		    }
		 } 
#endif
		 break;
	    case RUN_RECEIVE:
		if ((time(NULL)-timestamp)>=2UL) { /* 2 seconds passed? */
		    TesterPresent();     /* send Tester Present message */
		    time(&timestamp);    /* to keep modules in RUN mode */
		}
		if (!(modemstatus() & DSR)) {
		    printf("\n\nModule has reverted to INIT mode.\n");
		    mode=INIT_MODE;
		    break;
		}
		if (printmenu) {
		    printf("\nPress ESC to exit RECEIVE mode.\n");
		    printmenu=0;
		}
		while ((data=MyReadChar())>=0) {
		    printf("%02X ",data);
		    if (kbhit()) {
		        printmenu=1;
		        data=getch();
		        if (data==27) {
			    mode=RUN_MODE;
			    break;
			}
		        if (!data) getch(); /* clear extended code */
		    } 
		}
		break;
	    case RUN_RECEIVE_SEQ:
#ifdef NOT_NECESSARY
		if ((time(NULL)-timestamp)>=2UL) { /* 2 seconds passed? */
		    TesterPresent();     /* send Tester Present message */
		    time(&timestamp);    /* to keep modules in RUN mode */
		}
#endif
		if (!(modemstatus() & DSR)) {
		    printf("\n\nModule has reverted to INIT mode.\n");
		    mode=INIT_MODE;
		    break;
		}
#ifdef NOT_NECESSARY
		if (printmenu) {
		    printf("\nPress ESC to exit RECEIVE mode.\n");
		    printmenu=0;
		}
#endif
		while ((data=MyReadChar())>=0) {
		    if (seqstart) {
			if (data==incval) {
			    incval++;
                            /* if (!incval) putchar('.'); */
                            sound(data+500);
			} else {
			    printf("Expected %02X Got %02X.\n",incval,data);
			    printf("Returned to INIT mode.\n");
			    nosound();
			    mode=INIT_MODE;		
			    break;
			}
		    } else {
			seqstart=1;
			incval=data+1;
		    }	    
		    if (kbhit()) {
		        printmenu=1;
		        data=getch();
		        if (data==27) {
			    mode=RUN_MODE;
			    nosound();
			    break;
			}
		        if (!data) getch(); /* clear extended code */
		    } 
		}
		break;
	    case RUN_PARM:
		MyClearRXBuffer();
		if ((time(NULL)-timestamp)>=2UL) { /* 2 seconds passed? */
		    TesterPresent();   /* send Tester Present message */
		    time(&timestamp);    /* to keep modules in RUN mode */
		}
		if (!(modemstatus() & DSR)) {
		    printf("\n\nModule has reverted to INIT mode.\n");
		    mode=INIT_MODE;
		    break;
		}
		if (printmenu) {
		    printmenu=0;
		    stringpos=0;
		    printf("\nRUN MODE SELECT PARAMETER\n\n");
		    for (i=0; i<NUMBER_OF_PARAMETERS; i++) {
			printf("%u. %s\n",i+1,GetNamedParameter(i));
		    }
		    printf("%u. Quit to Main Run Mode Menu\n",
			NUMBER_OF_PARAMETERS+1);
		    printf("\nEnter Choice (1-%u): ",NUMBER_OF_PARAMETERS+1);
		} 
		if (kbhit()) {
		    string[stringpos]=getch();
		    switch (string[stringpos]) {
			case 0: getch(); break; /* special char - ignore */
			case 8:  /* backspace */
			    if (stringpos) {
				printf("\x08 \x08");
				stringpos--;
			    } break;
			case 13: case 10: /* carriage return */
			    string[stringpos]=0;
			    printf("\n");
			    printmenu=1;
			    ParmNum=atoi(string);
			    if ((ParmNum<1)||(ParmNum>(NUMBER_OF_PARAMETERS+1)))
			    {
				printf("\nIllegal Selection\n");
			    } else {
				if (ParmNum==(NUMBER_OF_PARAMETERS+1)) {
				    mode=RUN_MODE;
				} else {
				    ParmNum--;
				    mode=RUN_SET_PARM;
				}
			    }
			    break;
			default:
			    if (CANDISP(string[stringpos])) {
				printf("%c",string[stringpos]);
			        if (stringpos<50) stringpos++;
			    }
		    }
		 } break;
	    case RUN_SET_PARM:
		MyClearRXBuffer();
		if ((time(NULL)-timestamp)>=2UL) { /* 2 seconds passed? */
		    TesterPresent();   /* send Tester Present message */
		    time(&timestamp);    /* to keep modules in RUN mode */
		}
		if (!(modemstatus() & DSR)) {
		    printf("\n\nModule has reverted to INIT mode.\n");
		    mode=INIT_MODE;
		    break;
		}
		if (printmenu) {
		    printmenu=0;
		    stringpos=0;
		    printf("\nRUN MODE SET PARAMETER\n\n");
		    printf("Parameter %s currently = %02X\n\n"
			"Set Parameter %s = ",GetNamedParameter(ParmNum),
			Parameter[ParmNum],GetNamedParameter(ParmNum));
		} 
		if (kbhit()) {
		    string[stringpos]=getch();
		    switch (string[stringpos]) {
			case 0: getch(); break; /* special char - ignore */
			case 8:  /* backspace */
			    if (stringpos) {
				printf("\x08 \x08");
				stringpos--;
			    } break;
			case 13: case 10: /* carriage return */
			    string[stringpos]=0;
			    printf("\n");
			    if ((string[0]==0)||(string[0]==27)) {
			        printmenu=1;
				printf("\nAborted.\n");
				mode=RUN_MODE;
			    } else {
				if ((res=GetHex(&val))<0) {
			            printmenu=1;
				    printf("\nAborted.\n");
				    mode=RUN_MODE;
				} else {
				    if (res==0) {
				         stringpos=0;
				    } else {	
					Parameter[ParmNum]=val;
					SaveParameters();
					printf("\nParameter saved.\n");
					printmenu=1;
					mode=RUN_PARM;
				    }
				}
			    }
			    break;
			default:
			    if (CANDISP(string[stringpos])) {
				printf("%c",string[stringpos]);
			        if (stringpos<50) stringpos++;
			    }
		    }
		 } break;
	}
    }
    nosound();
    printf("\nTerminating program.\n");
    clean(0);
    return(0);
}

void clean(int error_code)
/* Close all open devices and return to OS */
{
    clean_rs232();
    exit(error_code);
}

void SendMessage(void)
{
unsigned char retry;
unsigned int resp;
unsigned char i,cks;
    Checksum=((unsigned char)(HEADER(2)+Target+Source+0x1A+0x80));
    printf("\nSending message %02X %02X %02X 1A 80 %02X...\n",
        HEADER(2),Target,Source,Checksum);
    if (mode==RUN_MODE) {
	MyWriteChar(HEADER(2));
	MyWriteChar(Target);
	MyWriteChar(Source);
	MyWriteChar(0x1A);
	MyWriteChar(0x80);
	MyWriteChar(Checksum);
    } else {
	ISOarg[0]='M';
	ISOarg[1]=6;
	ISOarg[2]=HEADER(2);
	ISOarg[3]=Target;
	ISOarg[4]=Source;
	ISOarg[5]=0x1A;
	ISOarg[6]=0x80;
	ISOarg[7]=Checksum;
        for (retry=1; retry<=5; retry++) {
	    SendMultiByteCommand(8);
	    if (RecvInitResp(&resp,RESP_STATUS)>=0) break;
        }
        if (retry<=5) {
	    if (resp=='O') {
	         printf("\nMessage Sent.\n");
	    } else {
	         printf("\nError Code = %02X (%s)\n", resp,GetErrorCode(resp));
	    }
	} else printf(NotResp);
    }
}
